/* file  cMunsell.cpp
 *
 * This file holds the implementation of my cMunsell class, which uses
 * Direct3D 8 to draw a three-dimensional model of the Munsell color solid.
 *
 * Because this code employs Direct3D, you must link with D3D8.LIB,
 * D3DX8.LIB, and DXERR8.LIB.
 *
 * You can check out my other software at:
 *   www.computersciencelab.com
 *
 *                                            John Kopplin  1/03
 */


///////////////////////////////////////////////////////////////////////////////
//  Includes
//
///////////////////////////////////////////////////////////////////////////////

#include "cMunsell.h"
#include "cMunsellResource.h"           // resource header file for cMunsell


///////////////////////////////////////////////////////////////////////////////
//  Defines and Typedefs
//
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//  Globals
//
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//  Externs
//
///////////////////////////////////////////////////////////////////////////////

extern IDirect3D8        * g_pD3D;
extern IDirect3DDevice8  * g_pd3dDevice;

extern CD3DFont          * g_pFont;   // font for writing on a Direct3D window



///////////////////////////////////////////////////////////////////////////////
// cMunsell::cMunsell()
//
///////////////////////////////////////////////////////////////////////////////

cMunsell::cMunsell()
    : m_FVF ( D3DFVF_XYZ        // we provide pre-transformed coordinates 
         // | D3DFVF_XYZRHW     // we provide transformed coordinates
         // | D3DFVF_NORMAL     // we provide a surface normal vector
         // | D3DFVF_DIFFUSE    // we specify a diffuse color for each vertex
         // | D3DFVF_SPECULAR   // we specify a specular color
            | D3DFVF_TEX1       // we employ 1 set of texture coordinates
         // | D3DFVF_TEXCOORDSIZE2( 0 )
                                // first set (index 0) of texture coordinates
                                //   has 2 elements (this declaration is not
                                //   necessary since this macro = 0)
            ),
      m_d3dcrBackground( D3DCOLOR_XRGB( 255, 255, 255 ) ), // white background
      m_nVertices( 0 ),
      m_nIndices( 0 ),
      m_pVertexBuffer( NULL ),
      m_pIndexBuffer( NULL ),
      m_vObjectCenter( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ) ),
      m_fObjectRadius( 0.0f ),
      m_fViewRadius( 2.0f ),
      m_fRotation( 0.0f ),
      m_fElevation( 0.5f ),
      m_vEyePt( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ) ),
      m_vLookatPt( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ) ),
                                  // we look at the origin
      m_vUpVector( D3DXVECTOR3( 0.0f, 0.0f, 1.0f ) ),
                                  // the z axis is the up direction
      m_bBackfaceCulling( false ),
      m_bTransparency( true ),
      m_bPaintersAlgorithm( true )
{
    UINT    i;

    for ( i=0; i < NSLICES; i++ )
        m_pTextures[i] = NULL;

} // cMunsell::cMunsell()


///////////////////////////////////////////////////////////////////////////////
// cMunsell::~cMunsell()
//
///////////////////////////////////////////////////////////////////////////////

cMunsell::~cMunsell()
{

} // cMunsell::~cMunsell()


///////////////////////////////////////////////////////////////////////////////
// cMunsell::CreateGeometry()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT cMunsell::CreateGeometry()
{
    // "Resource management" is the process in which resources are
    // promoted from system-memory storage to device-accessible
    // storage before rendering, then are discarded from
    // device-accessible storage to make room for other resources.
    // The resource manager for DirectX 8.0 manages textures as
    // before and now can manage vertex buffers and index buffers.
    // Use the D3DPOOL_MANAGED flag at creation time to specify
    // a managed resource. The Direct3D runtime then uses its own
    // management algorithm based on a least recently used priority
    // technique.  Managed resources persist through transitions
    // between the lost and operational states of the device. The
    // device can be restored with a call to IDirect3DDevice8::Reset()
    // and such resources continue to function normally without being
    // reloaded with artwork.  Resources in the default pool do NOT
    // persist through transitions between the lost and operational
    // states of the device. These resources must be released before
    // calling Reset() and must then be recreated.

    HRESULT    hr;
    IDirect3DDevice8 * pd3dDevice = g_pd3dDevice;

    // Set up our vertex buffer.

    // Each vertical slice of the Munsell color model is displayed via
    // a two-sided rectangle.  Luckily, we can simply turn off
    // back side culling and thereby achieve a two-sided rectangle
    // using only 4 vertices (2 triangles).

    // Each such rectangle requires 4 vertices and 6 indices.

    m_nVertices = NSLICES * 4;

    m_nIndices = NSLICES * 6;

    hr = pd3dDevice->CreateVertexBuffer(
                           sizeof(MYVERTEX) * m_nVertices,
                                              // bytes, not vertices
                           D3DUSAGE_WRITEONLY,
                                  // our app wont read from this array
                           m_FVF,
                           D3DPOOL_MANAGED,
                                  // type of memory management, in this case
                                  // we allow the device driver to determine
                                  // the most optimum memory location
                           &m_pVertexBuffer );

    if ( FAILED( hr ) )
        return hr;

    // The call to CreateVertexBuffer() increments the reference count
    // on the returned vertex buffer.  This reference count is decremented
    // when we call DIRECT3DVERTEXBUFFER8::Release() in ReleaseGeometry().

    // Set up an index buffer which is a list of indices into the
    // vertex array.

    hr = pd3dDevice->CreateIndexBuffer(
                                 sizeof(WORD) * m_nIndices,
                                              // bytes, not indices
                                 D3DUSAGE_WRITEONLY,
                                 D3DFMT_INDEX16,
                                    // since our indices[] holds WORDs
                                 D3DPOOL_MANAGED,
                                 &m_pIndexBuffer );

    if ( FAILED( hr ) )
        return hr;

    // The call to CreateIndexBuffer() increments the reference count
    // on the returned index buffer.  This reference count is decremented
    // when we call DIRECT3DINDEXBUFFER8::Release() in ReleaseGeometry().

    // We now need to fill our vertex buffer.

    // Locking a resource means granting CPU access to its storage.
    // Any other operations involving that resource are serialized
    // for the duration of a lock.  No accelerator operations on a
    // surface can be ongoing while a lock operation is outstanding
    // on that surface.

    // Lock() returns a pointer to where you can write your vertex data.

    // Note that you don't need to be concerned about a device
    // becoming lost during a lock operation.

    MYVERTEX   * pVertices;

    hr = m_pVertexBuffer->Lock( 0,                 // offset from the start
                                0,                 // lock the entire vertex buf
                                (BYTE**)&pVertices,
                                0 );               // flags

    if ( FAILED( hr ) )
        return hr;

    // Now we need to fill the index buffer.

    WORD       * pIndices;

    hr = m_pIndexBuffer->Lock( 0,                 // offset from the start
                               0,                 // lock the entire vertex buf
                               (BYTE **)&pIndices,
                               0 );               // flags

    if ( FAILED( hr ) )
    {
        m_pVertexBuffer->Unlock();
        return hr;
    }

    UINT    iSlice;

    for ( iSlice=0; iSlice < NSLICES; iSlice++ )
    {
        double   dAngle = iSlice * TWOPI / NSLICES;

        // The convention for texture coordinates is shown below:
        //
        //   (tu,tv) = (0,0)   ------------------   (w,0)
        //  upper left corner |                  |
        //                    |                  |
        //                    |                  |
        //                    |                  |
        //             (0,h)   ------------------   (w,h)

        pVertices[iSlice*4].x = 0.0f;
        pVertices[iSlice*4].y = 0.0f;
        pVertices[iSlice*4].z = -0.5f;
        pVertices[iSlice*4].tu = 0.05f;  // by starting at x = .05 we skip
                                         //   the first 8 of the 128 columns
                                         //   in each .BMP file used for texture
        pVertices[iSlice*4].tv = 1.0f;

        pVertices[iSlice*4+1].x = 0.0f;
        pVertices[iSlice*4+1].y = 0.0f;
        pVertices[iSlice*4+1].z = 0.5f;
        pVertices[iSlice*4+1].tu = 0.05f;
        pVertices[iSlice*4+1].tv = 0.0f;

        pVertices[iSlice*4+2].x = (float) cos( dAngle );
        pVertices[iSlice*4+2].y = (float) sin( dAngle );
        pVertices[iSlice*4+2].z = 0.5f;
        pVertices[iSlice*4+2].tu = 1.0f;
        pVertices[iSlice*4+2].tv = 0.0f;

        pVertices[iSlice*4+3].x = (float) cos( dAngle );
        pVertices[iSlice*4+3].y = (float) sin( dAngle );
        pVertices[iSlice*4+3].z = -0.5f;
        pVertices[iSlice*4+3].tu = 1.0f;
        pVertices[iSlice*4+3].tv = 1.0f;

        pIndices[iSlice*6+0] = iSlice*4 + 0;
        pIndices[iSlice*6+1] = iSlice*4 + 1;
        pIndices[iSlice*6+2] = iSlice*4 + 2;
        pIndices[iSlice*6+3] = iSlice*4 + 2;
        pIndices[iSlice*6+4] = iSlice*4 + 3;
        pIndices[iSlice*6+5] = iSlice*4 + 0;
    }

    // Compute m_fObjectRadius and m_vObjectCenter, which
    // are used for setting the D3DTS_PROJECTION matrix.

    // The vertex buffer has to be locked before we can call
    // D3DXComputeBoundingSphere() to generate a bounding sphere.

    hr = D3DXComputeBoundingSphere( pVertices,
                                    m_nVertices,
                                    m_FVF,
                                    &m_vObjectCenter,
                                    &m_fObjectRadius );

    hr = m_pVertexBuffer->Unlock();

    hr = m_pIndexBuffer->Unlock();

    if ( FAILED( hr ) )
        return hr;

    // We use a D3DX utility function to create an IDirect3DTexture8
    // interface. The D3DX image file loader functions support BMP,
    // TGA, PNG, JPG, DIB, PPM, and DDS files.

    D3DCOLOR   d3dcrTransparent;

    if ( m_bTransparency )
        d3dcrTransparent = D3DCOLOR_XRGB( 192, 192, 192 );
    else
        d3dcrTransparent = 0;

    UINT   i;

    for ( i=0; i < NSLICES; i++ )
    {
        hr = D3DXCreateTextureFromResourceEx(
                                      pd3dDevice,         // LPDIRECT3DDEVICE8
                                      NULL,               // HMODULE
                                      MAKEINTRESOURCE( IDB_Texture0 + i ),
                                      D3DX_DEFAULT,       // width
                                      D3DX_DEFAULT,       // height
                                      D3DX_DEFAULT,       // mip level
                                      0,                  // usage
                                       // If D3DUSAGE_RENDERTARGET is specified,
                                       // Pool must be set to D3DPOOL_DEFAULT
                                      D3DFMT_UNKNOWN,     // D3DFORMAT
                                      D3DPOOL_MANAGED,    // memory pool 
                                      D3DX_DEFAULT,       // filter
                                      D3DX_DEFAULT,       // mip filter
                                      d3dcrTransparent,   // D3DCOLOR intended
                                                          //  to be transparent
                                      NULL,               // D3DXIMAGE_INFO*
                                      NULL,               // PALETTEENTRY*
                                      &m_pTextures[i]);   // IDirect3DTexture8*
    }

    if ( FAILED( hr ) )
        return hr;

    // Store the render target's surface description.

    PDIRECT3DSURFACE8   pBackBuffer;
    D3DSURFACE_DESC     d3dsdBackBuffer;    // surface desc of the backbuffer

    pd3dDevice->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
    pBackBuffer->GetDesc( &d3dsdBackBuffer );
    pBackBuffer->Release();

    // Set up the world matrix to center the object at the origin.

    D3DXMATRIX     matWorld;

    D3DXMatrixIdentity( &matWorld );

    hr = pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    // Set up the projection matrix.

    D3DXMATRIX     matProj;
    float          fAspect = (float)d3dsdBackBuffer.Width /
                                              (float)d3dsdBackBuffer.Height;

    D3DXMatrixPerspectiveFovLH( &matProj,
                                D3DX_PI/4,
                                fAspect, 
                                m_fObjectRadius/64.0f,
                                m_fObjectRadius*100.0f );

    hr = pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

    return hr;

} // cMunsell::CreateGeometry()


///////////////////////////////////////////////////////////////////////////////
// cMunsell::ReleaseGeometry()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT cMunsell::ReleaseGeometry()
{
    HRESULT   hr;

    hr = g_pd3dDevice->SetTexture( 0, NULL );  // prevents a memory leak

    UINT    i;

    for ( i=0; i < NSLICES; i++ )
        SAFE_RELEASE( m_pTextures[i] );

    SAFE_RELEASE( m_pVertexBuffer );
    SAFE_RELEASE( m_pIndexBuffer );

    return hr;

} // cMunsell::ReleaseGeometry()


///////////////////////////////////////////////////////////////////////////////
// cMunsell::SetRenderStates()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT cMunsell::SetRenderStates()
{
    HRESULT     hr;
    IDirect3DDevice8 * pd3dDevice = g_pd3dDevice;

    // The D3DRENDERSTATETYPE enumerated type is used with
    // IDirect3DDevice8::SetRenderState() to specify all possible
    // rendering states.

    // By default, Direct3D performs lighting calculations on all
    // vertices.  And consequently, if a vertex has no normal vector
    // it will receive zero light.  An application that specifies
    // vertex colors probably will not specify normal vectors and
    // hence must be sure to disable lighting or else everything
    // will be black.

    // If you plan to cover all surfaces with textures then you
    // probably want to disable Direct3D lighting.  Otherwise,
    // you will use IDirect3DDevice8::SetLight() and
    // IDirect3DDevice8::SetMaterial().

    hr = pd3dDevice->SetRenderState( D3DRS_LIGHTING, false );

    if ( FAILED( hr ) )
        return hr;

    // Depth buffering is a method of removing hidden lines and surfaces.
    // By default, Direct3D does not use depth buffering but it can
    // be enabled using the D3DRS_ZENABLE state.

    // Enable the z-buffer.

    hr = pd3dDevice->SetRenderState( D3DRS_ZENABLE, D3DZB_TRUE );

    if ( FAILED( hr ) )
        return hr;

    // To convince Direct3D to render the inside of an object in
    // addition to its outside you will need to turn off back-face
    // culling.

    hr = pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    if ( FAILED( hr ) )
        return hr;

    // Since we want to employ color keying when we load our textures,
    // we need to enable alpha blending.

    hr = pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, true );

    if ( FAILED( hr ) )
        return hr;

    // Alpha blending means combining the color value of the new pixel
    // with the pixel already stored at that location in the frame buffer.
    // There are a number of ways of doing this, described by the eq.:

    // FinalColor = TexelColor  SourceBlendFactor
    //            + PixelColor  DestBlendFactor

    // To achieve complete transparency you would request:

    // hr = pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ZERO ); 
    // hr = pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

    // But we want the alpha blending formula to vary on a pixel by pixel
    // basis, as described by the alpha channel of the source.  This
    // is accomplished by:

    hr = pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA ); 

    if ( FAILED( hr ) )
        return hr;

    hr = pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

    if ( FAILED( hr ) )
        return hr;

    // When I originally wrote this program, the transparent
    // textures that I use for the vertical slices through the
    // Munsell color solid were interfering with each other
    // causing a distorted picture.  I discovered that I could
    // solve the problem by being careful to always render the
    // slices in a back-to-front order (the "painter's algorithm")
    // whereas my original code always rendered slice #0 first,
    // then slice #1, etc.  You can observe this solution in the
    // cMunsell::Render() function where I select the slice
    // rendering order based upon the current value of m_fRotation.
    // I was motivated to publish this code as a Direct3D
    // tutorial in order to demonstrate the importance of the
    // painter's algorithm, even in a program that employs
    // transparent textures.

    // But there is an easier method that was taught to me by
    // Henrik Rydgrd of Sweden.  It turns out that I CAN
    // employ my original render algorithm (i.e., always start
    // with slice #0 regardless of m_fRotation) if I configure
    // Direct3D as shown below:

//  pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE, true );
//  pd3dDevice->SetRenderState( D3DRS_ALPHAREF, 0x01 ); // alpha value to
                                                        //  compare against
//  pd3dDevice->SetRenderState( D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL );

    // These statements ask Direct3D to NOT write a transparent
    // pixel to the z buffer unless its alpha value exceeds a
    // certain value (the value 0x01 is sufficient for my textures).
    // If the transparent pixels from slice #0 are not written
    // to the depth buffer then the (non-transparent) pixels
    // from later slices can still qualify for rendering and
    // thereby appear to be UNDERNEATH the slice that was drawn
    // earlier.

    // I have left these 3 statements commented out because many
    // people's video cards might not support this capability.
    // You can interrogate your video card to learn if it has this
    // capability in the following manner:

//  D3DCAPS8   d3dCaps;

//  pd3dDevice->GetDeviceCaps( &d3dCaps );

//  if ( d3dCaps.AlphaCmpCaps & D3DPCMPCAPS_GREATEREQUAL )
//  {
//      pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE, true );
//      pd3dDevice->SetRenderState( D3DRS_ALPHAREF, 0x01 ); // alpha value to
                                                        //  compare against
//      pd3dDevice->SetRenderState( D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL );
//  }
//  else
//  {
//      do something else
//  }

    // Because of this hardware dependency, it is safer for
    // this program to render the slices in a back-to-front order
    // because this will work with all video cards.  You can use
    // the 'p' keystroke to alternate between the original (dumb)
    // render algorithm and the smarter back-to-front algorithm.
    // If you uncomment the above 3 statements (and your video
    // card supports this feature) then you will see that even the
    // original dumb algorithm now works fine.

    return hr;

} // cMunsell::SetRenderStates()


///////////////////////////////////////////////////////////////////////////////
//  cMunsell::Render()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT cMunsell::Render()
{
    HRESULT     hr;

    if ( m_bBackfaceCulling )
        hr = g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
    else
        hr = g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    // A depth buffer (a.k.a. z-buffer) is a memory area used to
    // determine how the pixels of rasterized polygons occlude one
    // another. When a 3-D scene is rasterized with depth buffering
    // enabled, each point on the rendering surface is tested.
    // At the beginning of the test, the depth value in the depth
    // buffer is set to the largest possible value for the scene.
    // The color value on the rendering surface is set to either
    // the background color value or the color value of the
    // background texture at that point. Each polygon in the scene
    // is tested to see if it intersects with the current coordinate
    // (x,y) on the rendering surface. If it does, the depth value --
    // which will be the z coordinate in a z-buffer or the w
    // coordinate in a w-buffer -- at the current point is tested
    // to see if it is smaller than the depth value stored in the
    // depth buffer. If the depth of the polygon value is smaller,
    // it is stored in the depth buffer and the color value from
    // the polygon is written to the current point on the rendering
    // surface.

    // Nearly all 3-D accelerators on the market support z-buffering,
    // making z-buffers the most common type of depth buffer today.
    // However the generated z values in a z-buffer tend not to be
    // distributed evenly across the z-buffer range (typically 0.0
    // to 1.0, inclusive).  For example, using a far-plane distance to
    // near-plane distance ratio of 100, 90% of the depth buffer
    // range is spent on the first 10% of the scene depth range.
    // A w buffer is often more evenly distributed between the
    // near and far clip planes than a z-buffer but w buffering
    // isn't supported as widely in hardware.

    // Before rendering objects in a scene, clear the viewport on the
    // render-target surface.  Applications that render scenes covering
    // the entire area of the render-target surface can improve
    // performance by clearing the attached depth and stencil buffer
    // surfaces, if any, instead of the render target. However, if your
    // application renders only to a portion of the render-target
    // surface, explicit clear operations are required.

    // IDirect3DDevice8::Clear() flood fills the buffers you specify.

    IDirect3DDevice8 * pd3dDevice = g_pd3dDevice;

    hr = pd3dDevice->Clear( 0,               // count of rects
                            NULL,            // ptr to array of rects
                            D3DCLEAR_TARGET  // fill the back buffer...
                          | D3DCLEAR_ZBUFFER,
                            m_d3dcrBackground,
                            1.0,             // fill value for z buffer
                            0 );
    if ( FAILED( hr ) )
        return hr;

    // Rather than rotating the object, we rotate the camera (eye).
    // The advantage is that then the consequence of pitch, roll,
    // and yaw rotations remain predictable since the world
    // coordinate axes maintain a constant relationship to the
    // object.  The code that sets the D3DTS_WORLD matrix is found
    // in CreateGeometry() while the code that sets the
    // D3DTS_VIEW matrix is found in Render().

    D3DXMATRIX     matView;

    m_vEyePt.x = (float)( m_fViewRadius * cos( m_fRotation ) * cos( m_fElevation ) );
    m_vEyePt.y = (float)( m_fViewRadius * sin( m_fRotation ) * cos( m_fElevation ) );
    m_vEyePt.z = (float)( m_fViewRadius * sin( m_fElevation ) );

    D3DXMatrixLookAtLH( &matView, &m_vEyePt, &m_vLookatPt, &m_vUpVector );

    hr = pd3dDevice->SetTransform( D3DTS_VIEW,  &matView );

    if ( FAILED( hr ) )
        return hr;

    // The World transformation moves the object into world coordinates.
    // The View transformations move the world into view space.
    // The Projection matrix scales the world to make it look as if
    // it has depth.

    // A model's vertices are declared relative to a local coordinate system
    // called "model space". The first stage of the geometry pipeline
    // transforms a model's vertices from their local coordinate system
    // to a coordinate system that is used by all the objects in a scene.
    // The process of reorienting the vertices is called the "world
    // transformation". This new orientation is commonly referred to as
    // "world space", and each vertex in world space is declared using
    // world coordinates. In the next stage, the vertices that describe
    // your 3-D world are oriented with respect to a camera. That is,
    // your application chooses a point-of-view for the scene, and world
    // space coordinates are relocated and rotated around the camera's view,
    // turning world space into "camera space". This is the "view
    // transformation". In camera space, the camera, or viewer, is at
    // the origin, looking in the positive z-direction.
    // The next stage is the projection transformation. This
    // stage is analogous to choosing a lens for the camera.
    // In this part of the pipeline, objects are usually scaled with
    // relation to their distance from the viewer in order to give the
    // illusion of depth to a scene. In the final part of the pipeline,
    // any vertices that will not be visible on the screen are clipped.
    // After clipping, the remaining vertices are scaled according to
    // the viewport parameters and converted into screen coordinates.
    // The resulting vertices -- seen on the screen when the scene is
    // rasterized -- exist in screen space.

    // Start drawing.

    // You can only call rendering primitive functions after
    // you have called IDirect3DDevice8::BeginScene() to notify
    // Direct3D that rendering is to begin.  All rendering methods
    // must be bracketed by calls to BeginScene() and EndScene().
    // Do NOT call GDI functions between BeginScene() and EndScene().

    hr = pd3dDevice->BeginScene();

    if ( FAILED( hr ) )
        return hr;

    // If BeginScene() returns any error, you must not call
    // EndScene() because the scene has not successfully begun. 
    // If BeginScene() succeeds, but you get an error while rendering
    // the scene, you must call EndScene().  If EndScene() fails,
    // for any reason, you need not call it again. 

    // Specify the active index buffer.

    hr = pd3dDevice->SetIndices( m_pIndexBuffer,
                                 0 );        // base vertex index

//    if ( FAILED( hr ) )
//        return hr;
// WRONG because it would prevent the necessary call to EndScene()

    // Specify a fixed function vertex shader using the mesh FVF.

    hr = pd3dDevice->SetVertexShader( m_FVF );

//    if ( FAILED( hr ) )
//        return hr;
// WRONG because it would prevent the necessary call to EndScene()

    // Specify the active vertex data stream (you are allowed multiple
    // vertex arrays so you have to specify which vertex data stream is
    // active).

    hr = pd3dDevice->SetStreamSource( 0,                  // stream #
                                      m_pVertexBuffer,
                                      D3DXGetFVFVertexSize( m_FVF ) );
                                                          // stride

    // WARNING: In Direct3D 9 an additional offset parameter has
    // been added to the SetStreamSource() function.

//    if ( FAILED( hr ) )
//        return hr;
// WRONG because it would prevent the necessary call to EndScene()

    // Now that we have set the stream source and vertex shader, any
    // draw methods will use the vertex buffer.

if ( m_bPaintersAlgorithm )
{
    // This section of code is explained at:
    //    http://www.computersciencelab.com/Direct3DTut1.htm

    // The painter's algorithm indicates that we have to be careful
    // to render those objects furthest from the viewer before we
    // render closer objects.  Therefore we must paint the vertical
    // slices of the Munsell color solid in an order that depends
    // upon the current rotation of the model.  And we have to do
    // this in two steps, by considering each half of the cylindrical
    // color model. We must draw the vertical slices residing on the
    // left hand of the Munsell color model from the back-most slice
    // to the front-most slice and then we must draw the vertical slices
    // residing on the right hand side starting with the back-most
    // slice and proceeding to the front-most slice.
    
    // This is summarized in the table below which assumes NSLICES=20.
    // Note that m_fRotation varies from 0 to TWOPI but this table
    // shows it varying from 0 to 360 degrees.
    //
    //  m_fRotation    SlicesDrawOnLeft     SlicesDrawnOnRight
    //    0 to 18        10 to 1              11 to 19 and 0
    //   18 to 36        11 to 2              12 to 19 and 0 to 1
    //   36 to 54        12 to 3              13 to 19 and 0 to 2
    //   54 to 72        13 to 4              14 to 19 and 0 to 3
    //   72 to 90        14 to 5              15 to 19 and 0 to 4
    //     ...
    //  144 to 162       18 to 9              19 and 0 to 8
    //  162 to 180       19 to 10              0 to  9
    //  180 to 198       0 and 19 to 11        1 to 10
    //  198 to 216       1 to 0 and 19 to 12   2 to 11
    //     ...
    //  306 to 324       7 to 0 and 19 to 18   8 to 17
    //  324 to 342       8 to 0 and 19         9 to 18
    //  342 to 359       9 to 0               10 to 19

    double  dRotationPerSlice = TWOPI / NSLICES;      // in rads

    UINT    uiRotationInSlices = (UINT)(m_fRotation / dRotationPerSlice);
                                            // we do want truncation here

    int     iFirstSliceDrawnOnLeft =
                                   ( NSLICES/2 + uiRotationInSlices ) % NSLICES;

    int     iFirstSliceDrawnOnRight = ( iFirstSliceDrawnOnLeft + 1 ) % NSLICES;

    // The following for() loop paints the slices visible on the
    // left-hand side of the Munsell color model.

    int    iLastDrawn;

    if ( iFirstSliceDrawnOnLeft >= NSLICES / 2 )
        iLastDrawn = iFirstSliceDrawnOnLeft - (NSLICES/2 - 1);
    else
        iLastDrawn = 0;

    int    iSlice;
    UINT   nDrawnFirstPass = 0;
    UINT   nDrawnSecondPass = 0;

    for ( iSlice = iFirstSliceDrawnOnLeft; iSlice >= iLastDrawn; iSlice-- )
    {
        // Specify a texture for stage zero.  Direct3D 8
        // supports multitexturing which allows you to have up
        // to eight stages of textures which are blended together.
        // The .MaxSimultaneousTextures member of the D3DCAPS8
        // struct returned by IDirect3DDevice8::GetDeviceCaps()
        // tells you how many textures are supported by the current
        // video card.

        hr = pd3dDevice->SetTexture( 0,      // texture stage identifier (0-7)
                                     m_pTextures[iSlice] );

//      if ( FAILED( hr ) )
//          return hr;
// WRONG because it would prevent the necessary call to EndScene()

        // The SetTexture() method increments the reference count of
        // the texture surface being assigned. When the texture is no
        // longer needed, you should set the texture at the appropriate
        // stage to NULL. If you fail to do this, the surface will not
        // be released, resulting in a memory leak.

        // Generally a texture is mapped onto a primitive image that is
        // larger or smaller than itself. If you do not specify a type
        // of texture filtering, Direct3D uses a technique called
        // nearest-point sampling (D3DTFP_POINT) which is computationally
        // simple but can result in visual distortion.  Direct3D provides
        // three types of texture filtering: linear filtering,
        // anisotropic filtering, and mipmap filtering; which -- at the
        // cost of additional computation time -- improve the mapping of
        // the texture to the primitive.  The majority of hardware today is
        // optimized for linear filtering, so your application may not
        // even suffer degraded performance when avoiding nearest-point
        // sampling.  Direct3D uses a form of linear texture filtering
        // (D3DTFP_LINEAR) called "bilinear filtering".  Bilinear filtering
        // finds the texel whose integer address is closest to the computed
        // address and then computes a weighted average of the texels that
        // are immediately above, below, to the left of, and to the right of
        // the nearest sample point.  A "mipmap" (D3DTSS_MIPFILTER) is a
        // sequence of textures, each of which is a progressively lower
        // resolution representation of the same image. The height and width
        // of each image, or level, in the mipmap is a power of two smaller
        // than the previous level. Mipmaps do not have to be square.

        // Enabling texture wrapping in one direction causes the system to
        // treat a texture as though it were wrapped around a cylinder.
        // Wrapping in both the u- and v-directions treats the texture
        // as being on the surface of a torus or doughnut. 

        hr = pd3dDevice->DrawIndexedPrimitive(
                                         D3DPT_TRIANGLELIST,
                                         iSlice*4, // minimum vertex index for
                                                   //   vertices used during
                                                   //   this call
                                         4,         // # of vertices
                                         iSlice*6,  // index number to start on
                                         2);        // # triangles to render
        nDrawnFirstPass++;

//    if ( FAILED( hr ) )
//        return hr;
// WRONG because it would prevent the necessary call to EndScene()
    }

    if ( nDrawnFirstPass < NSLICES/2 )
    {
        for ( iSlice = NSLICES-1, nDrawnSecondPass = 0;
                 nDrawnSecondPass < (NSLICES/2 - nDrawnFirstPass );
                    iSlice-- )
        {
            hr = pd3dDevice->SetTexture( 0,
                                         m_pTextures[iSlice] );

            hr = pd3dDevice->DrawIndexedPrimitive(
                                         D3DPT_TRIANGLELIST,
                                         iSlice*4, // minimum vertex index for
                                                   //   vertices used during
                                                   //   this call
                                         4,         // # of vertices
                                         iSlice*6,  // index number to start on
                                         2);        // # triangles to render
            nDrawnSecondPass++;
        }
    }

    // The following for() loop paints the slices visible on the
    // right-hand side of the Munsell color model.

    if ( iFirstSliceDrawnOnRight > NSLICES / 2 )
        iLastDrawn = NSLICES-1;
    else
        iLastDrawn = iFirstSliceDrawnOnRight + NSLICES/2 - 1;

    nDrawnFirstPass = 0;

    for ( iSlice = iFirstSliceDrawnOnRight; iSlice <= iLastDrawn; iSlice++ )
    {
        hr = pd3dDevice->SetTexture( 0,      // texture stage identifier (0-7)
                                     m_pTextures[iSlice] );

        hr = pd3dDevice->DrawIndexedPrimitive(
                                         D3DPT_TRIANGLELIST,
                                         iSlice*4, // minimum vertex index for
                                                   //   vertices used during
                                                   //   this call
                                         4,         // # of vertices
                                         iSlice*6,  // index number to start on
                                         2);        // # triangles to render
        nDrawnFirstPass++;
    }

    if ( nDrawnFirstPass < NSLICES/2 )
    {
        for ( iSlice = 0, nDrawnSecondPass = 0;
                 nDrawnSecondPass < (NSLICES/2 - nDrawnFirstPass);
                    iSlice++ )
        {
            hr = pd3dDevice->SetTexture( 0,
                                         m_pTextures[iSlice] );

            hr = pd3dDevice->DrawIndexedPrimitive(
                                         D3DPT_TRIANGLELIST,
                                         iSlice*4, // minimum vertex index for
                                                   //   vertices used during
                                                   //   this call
                                         4,         // # of vertices
                                         iSlice*6,  // index number to start on
                                         2);        // # triangles to render
            nDrawnSecondPass++;
        }
    }
}
else                    // don't use painter's algorithm, instead always
{                       //   render slice #0 first, then slice #1, etc.
    int   iSlice;

    for ( iSlice = 0; iSlice < NSLICES; iSlice++ )
    {
        hr = pd3dDevice->SetTexture( 0,      // texture stage identifier (0-7)
                                     m_pTextures[iSlice] );

        hr = pd3dDevice->DrawIndexedPrimitive(
                                         D3DPT_TRIANGLELIST,
                                         iSlice*4, // minimum vertex index for
                                                   //   vertices used during
                                                   //   this call
                                         4,         // # of vertices
                                         iSlice*6,  // index number to start on
                                         2);        // # triangles to render
    }
} // end of if ( m_bPaintersAlgorithm

    g_pFont->DrawText( 2,   0,            // x,y
                       D3DCOLOR_ARGB(255,0,255,0),  // fully opaque green
                       m_bBackfaceCulling ?
                        TEXT( "b to toggle backface culling, currently TRUE" )
                       :
                        TEXT( "b to toggle backface culling, currently FALSE" ) );

    g_pFont->DrawText( 2,  20,            // x,y
                       D3DCOLOR_ARGB(255,0,255,0),  // fully opaque green
                       m_bTransparency ?
                        TEXT( "t to toggle transparency, currently TRUE" )
                       :
                        TEXT( "t to toggle transparency, currently FALSE" ) );

    g_pFont->DrawText( 2,  40,            // x,y
                       D3DCOLOR_ARGB(255,0,255,0),  // fully opaque green
                       m_bPaintersAlgorithm ?
                        TEXT( "p to toggle painter's algorithm, currently TRUE" )
                       :
                        TEXT( "p to toggle painter's algorithm, currently FALSE" ) );

    g_pFont->DrawText( 2,  60,            // x,y
                       D3DCOLOR_ARGB(255,0,255,0),  // fully opaque green
                       TEXT( "r/R to rotate" ) );

    g_pFont->DrawText( 2, 80,             // x,y
                       D3DCOLOR_ARGB(255,0,255,0),
                       TEXT( "e/E to elevate" ) );

    g_pFont->DrawText( 2, 100,             // x,y
                       D3DCOLOR_ARGB(255,0,255,0),
                       TEXT( "z/Z to zoom" ) );

    hr = pd3dDevice->EndScene();

    if ( FAILED( hr ) )
        return hr;

    // The new IDirect3DDevice8::Present() method replaces the
    // Flip() and Blt() methods used in DirectX 7.  This provides
    // consistency for both windowed and full screen rendering
    // [however, internally a full-screen swap chain will typically
    // present images with the flipping DDI while a windowed swap chain
    // will present images with the blitting DDI].
    
    // Another improvement in Direct3D 8 is that the number of situations
    // in which an application will encounter loss of resources (and the
    // actions to take in such cases) have been reduced.

    // When a full screen device is created, the Direct3D object
    // that created the device is marked as owning all adapters
    // on that system. This state is known as exclusive mode and
    // it means that devices created by any other Direct3D object
    // can neither assume full-screen operations nor allocate
    // video memory. When a Direct3D object assumes exclusive mode,
    // all devices other than the one that went full-screen are
    // placed in lost state.  These other devices cannot be reset
    // and remain in lost state until the exclusive mode is lost.

    // Present() cycles to the next back buffer. Since we only have
    // one back buffer and one front buffer, the buffers simply flip.

    // The front buffer is not directly exposed in the Direct3D API
    // for DirectX 8.0. As a result, applications cannot lock or
    // render to the front buffer. 

    hr = pd3dDevice->Present( NULL, NULL, NULL, NULL );

    // The first two parameters passed to IDirect3DDevice8::Present()
    // are a source rectangle and destination rectangle.  Using
    // NULL for both these first two parameters causes the
    // the entire back buffer to be copied to the front buffer.
    // The third parameter sets the destination window for this
    // presentation. If you set this parameter to NULL, the
    // .hWndDeviceWindow member of D3DPRESENT_PARAMETERS is used.
    // And if the .hWndDeviceWindow member of D3DPRESENT_PARAMETERS
    // is NULL then the hFocusWindow param passed to
    // IDirect3D::CreateDevice() is used.
    // The fourth parameter is the DirtyRegion parameter used for
    // sub-rectangle updates; in most cases, it should be set to NULL.

    // A Direct3D device makes a transition to the "lost state" when
    // an event causes rendering to become impossible.  Examples of
    // such events are the loss of keyboard focus in a full-screen
    // application, or when the user presses ALT+TAB, or when a
    // system dialog is initialized, or when another application
    // assumes full-screen operation.

    // The application can continue to call rendering methods such
    // as IDirect3DDevice8::DrawPrimitive() without receiving
    // notification of a lost device. Internally, these operations
    // are discarded until the device is reset to the operational state.
    // IDirect3DDevice8::Reset() is the only method that has an effect
    // when a device is lost, and it is the only method by which an
    // application can change the device from a lost to an operational
    // state.
    
    // But before calling IDirect3DDevice8::Reset() the application
    // must destroy all video-memory resources and any swap chains.
    // Reset() will fail unless the application releases all resources
    // that are allocated in D3DPOOL_DEFAULT, including those created
    // by IDirect3DDevice8::CreateVertexBuffer() and
    // IDirect3DDevice8::CreateIndexBuffer().  Following the call to
    // Reset(), the application basically repeats its initialization
    // of Direct3D including a call to IDirect3DDevice8::CreateDevice()
    // which recreates the device.

    // The application's strategy for responding to a lost device
    // depends upon the value returned by
    // IDirect3DDevice8::TestCooperativeLevel().

    return hr;

} // cMunsell::Render()


///////////////////////////////////////////////////////////////////////////////
//  cMunsell::ToggleTransparency()
//
///////////////////////////////////////////////////////////////////////////////

void  cMunsell::ToggleTransparency()
{
    if ( m_bTransparency )
        m_bTransparency = false;
    else
        m_bTransparency = true;

    // We now need to re-read the .BMP files holding the various textures.

    HRESULT    hr;

    hr = ReleaseGeometry();      // releases vertex and index buffers
    hr = CreateGeometry();       // creates vertex and index buffers

    return;

} // cMunsell::ToggleTransparency()
